from surety.diff import compare
from surety.db.client import DbClient, ConnectionClient, to_dict
from tests.client.fake import FakeConnection, FakeConnectionNoRows


def test_dbclient_singleton():
    class Fake:
        def connect(self):
            return self

    c1 = DbClient('x', Fake())
    c2 = DbClient('x', Fake())

    assert c1 is c2


def test_to_dict_with_table():
    class Column:
        def __init__(self, name):
            self.name = name

    class Table:
        columns = [
            Column('id'),
            Column('global'),
            Column('metadata'),
            Column('empty')
        ]

    class Record:
        __table__ = Table
        id = 1
        is_global = True
        metadata_column = 'test'
        empty = None

    result = to_dict(Record(), filter_none=True)
    compare(
        actual=result,
        expected={'id': 1, 'is_global': True, 'metadata_column': 'test'},
    )


def test_execute_with_args():
    client = ConnectionClient()
    client.connection = FakeConnectionNoRows()

    client.execute('SQL', 1)


def test_execute_results():
    client = ConnectionClient()
    client.connection = FakeConnection()

    result = client.execute('SELECT 1')
    assert result == [(1,)]



from surety.db.client import PostgresqlDBClient
from tests.client.fake import FakeConnectionNoRows, FakeSession


def test_postgres_connect(monkeypatch):
    monkeypatch.setattr(
        'surety.db.client.create_engine',
        lambda url: object
    )
    monkeypatch.setattr(
        'surety.db.client.sessionmaker',
        lambda bind: FakeConnectionNoRows
    )

    client = PostgresqlDBClient(password='x')
    client.connect()

    assert client.connection is not None


def test_postgres_truncate():
    client = PostgresqlDBClient(password='pwd')
    client.execute = lambda sql: [('TRUNCATE other.users;',)]
    client.connection = object()
    client.trunc_all_tables(schemas=['other'], exclude_tables=['x'])



def test_postgres_sync_commit_true(monkeypatch):
    monkeypatch.setattr(
        'surety.db.client.create_engine',
        lambda url: object
    )
    monkeypatch.setattr(
        'surety.db.client.sessionmaker',
        lambda bind: FakeSession
    )

    client = PostgresqlDBClient(password='pwd', sync_commit=True)
    client.connect()



import pytest

from surety.db.client import BaseClient
from tests.data import Model
from tests.client.fake import FakeConnectionNoRows, FakeSession


def test_insert(monkeypatch):
    client = BaseClient()
    client.connection = FakeSession()

    class BaseModel:
        is_custom = False

        def __init__(self, *args, **kwargs):
            pass

        def to_table_record(self):
            return self

        def with_values(self, _):
            return self

    # 🔥 mock both to_dict and make_transient
    monkeypatch.setattr('surety.db.client.to_dict', lambda x: {})
    monkeypatch.setattr('surety.db.client.make_transient', lambda x: None)

    result = client.insert(BaseModel())

    assert result is not None


def test_get_records():
    client = BaseClient()
    client.connection = FakeSession()

    result = client.get_records(object, id=1)
    assert result == [1, 2, 3, 4]


def test_update():
    client = BaseClient()
    client.connection = FakeSession()
    client.update(Model, {'a': 1}, id=1)


def test_delete():
    client = BaseClient()
    client.connection = FakeSession()
    client.delete(Model, id=1)


def test_trunc_not_implemented():
    client = BaseClient()

    with pytest.raises(NotImplementedError):
        client.trunc_all_tables()



def test_cascade_delete():
    client = BaseClient()
    client.connection = FakeConnectionNoRows()

    class M:
        __table__ = type('T', (), {'__tablename__': 'users'})

    client.cascade_delete(M)
    assert 'TRUNCATE users CASCADE' in client.connection.sql


def test_get_records_with_criteria():
    client = BaseClient()
    client.connection = FakeSession()

    result = client.get_records(object, 'x > 1')
    assert result == [1, 2, 3, 4]


def test_update_with_criteria():
    client = BaseClient()
    client.connection = FakeSession()

    client.update(Model, {'a': 1}, 'id>1')


def test_delete_with_criteria():
    client = BaseClient()
    client.connection = FakeSession()

    client.delete(Model, 'id>1')


def test_pre_insert_and_commit():
    client = BaseClient()
    client.connection = FakeSession()

    client.pre_insert(object())
    client.commit_and_close()



import pytest

from surety.db.client import CassandraDBClient

class Query:
    def allow_filtering(self):
        return self

    def filter(self, **__):
        return ['filtered']

    def all(self):
        return ['filtered', 'non-filtered']


class CassandraModel:
    objects = Query()


def test_cassandra_get_records_kwargs():
    client = CassandraDBClient()
    result = client.get_records(CassandraModel, id=1)

    assert result == ['filtered']


def test_cassandra_get_records_all():
    client = CassandraDBClient()
    result = client.get_records(CassandraModel)

    assert result == ['filtered', 'non-filtered']


def test_cassandra_insert():
    class Table:
        @staticmethod
        def create(**kwargs):
            pass

    class Model:
        __table__ = Table

        def to_db(self):
            return {'a': 1}

    client = CassandraDBClient()
    client.connection = object()

    result = client.insert(Model())
    assert result is not None



def test_cassandra_truncate_loop():
    class FakeConn:
        def execute(self, *_, **__):
            return [{"table_name": "users"}]

    client = CassandraDBClient()
    client.connection = FakeConn()

    client.trunc_all_tables(exclude_tables=[])

def test_cassandra_connect_initializes(monkeypatch):
    setup_called = {}

    class FakeConnectionModule:
        def setup(self, hosts, keyspace, protocol_version=None, lazy_connect=None):
            setup_called["hosts"] = hosts
            setup_called["keyspace"] = keyspace
            setup_called["protocol_version"] = protocol_version
            setup_called["lazy_connect"] = lazy_connect

    # Patch the module path used in the inline import
    monkeypatch.setattr(
        "cassandra.cqlengine.connection",
        FakeConnectionModule(),
        raising=False
    )

    client = CassandraDBClient(
        hosts=["h1"],
        keyspace="ks",
        protocol_version=4
    )

    result = client.connect()

    assert result is client
    assert client.connection is not None
    assert setup_called["hosts"] == ["h1"]
    assert setup_called["keyspace"] == "ks"
    assert setup_called["protocol_version"] == 4
    assert setup_called["lazy_connect"] is True


def test_cassandra_cascade_delete_not_implemented():
    client = CassandraDBClient()

    with pytest.raises(NotImplementedError):
        client.cascade_delete(object)


def test_cassandra_update_not_implemented():
    client = CassandraDBClient()

    with pytest.raises(NotImplementedError):
        client.update(object, {})


def test_cassandra_delete_not_implemented():
    client = CassandraDBClient()

    with pytest.raises(NotImplementedError):
        client.delete(object)



class FakeQuery:
    def __init__(self):
        self.filtered = False

    def filter(self, *_):
        return self

    def filter_by(self, **__):
        return self

    def order_by(self, _):
        return self

    def update(self, *args, **kwargs):
        pass

    def delete(self, **kwargs):
        pass

    def all(self):
        return [1, 2, 3, 4]


class FakeSession:
    def __init__(self):
        self.added = None

    def add(self, obj):
        self.added = obj

    def commit(self):
        pass
    def refresh(self, obj):
        pass
    def close_all(self):
        pass
    def close(self):
        pass
    def flush(self):
        pass

    def query(self, _):
        return FakeQuery()


class FakeResultNoRows:
    returns_rows = False


class FakeConnectionNoRows:
    def commit(self):
        pass

    def close(self):
        pass

    def close_all(self):
        pass

    def execute(self, sql, *_, **__):
        self.sql = sql  # pylint: disable=attribute-defined-outside-init
        return FakeResultNoRows()


class FakeResult:
    returns_rows = True

    def fetchall(self):
        return [(1,)]


class FakeConnection:
    def execute(self, *_, **__):
        return FakeResult()



import json
import uuid
from datetime import datetime

from pytz import timezone

from surety import Int, String
from surety.diff import compare
from surety.db.types.common import DbTimestamp
from surety.db.types.mysql import DbBool, DbJsonArray as MysqlArray
from surety.db.types.postgres import DbUuid
from surety.db.types.cassandra import (
    DBJsonDict as CassandraDict, DBJsonArray as CassandraArray
)
from tests.data import SampleDict


DATE_VALUE = datetime.now().replace(
    year=2022, month=5, day=13, hour=12, minute=0, second=10, microsecond=0
)
TZ = datetime.now().replace(
    year=2022, month=5, day=13, hour=12, minute=0, second=10, microsecond=0
).astimezone(tz=timezone('CET'))
DATE = DbTimestamp().with_values(DATE_VALUE)
Uuid = DbUuid()


def test_to_db():
    assert DATE.to_db() == DATE_VALUE


def test_to_timezone():
    assert DATE.to_timezone('CET') == TZ


def test_to_text():
    assert DATE.to_format() == '2022-05-13'


def test_mysql_bool_with_values():
    assert DbBool().with_values(1).value is True


def test_mysql_bool_to_db():
    assert DbBool().with_values(0).to_db() == 0


def test_uuid():
    assert isinstance(DbUuid().with_values(uuid.uuid4()).value, str)


def test_json_dict_with_values():
    compare(
        actual=SampleDict().with_values('{"id":234,"name":"val"}').value,
        expected={
            SampleDict.Id.name: 234,
            SampleDict.Name.name: 'val'
        }
    )


def test_json_dict_to_db():
    assert SampleDict().with_values({
        SampleDict.Id.name: 1,
        SampleDict.Name.name: 'test'
    }).with_values(None).to_db() == '{"id":1,"name":"test"}'


def test_mysql_json_array_to_db():
    assert MysqlArray(SampleDict).with_values([{
        SampleDict.Id.name: 1,
        SampleDict.Name.name: 'test'
    }]).with_values(None).to_db() == '[{"id":1,"name":"test"}]'


def test_mysql_json_array_with_values():
    compare(
        actual=MysqlArray(SampleDict).with_values(
            '[{"id":234,"name":"val"}]'
        ).value,
        expected=[{
            SampleDict.Id.name: 234,
            SampleDict.Name.name: 'val'
        }]
    )


def test_cassandra_json_array_to_db():
    assert CassandraArray(SampleDict).with_values([{
        SampleDict.Id.name: 1,
        SampleDict.Name.name: 'test'
    }]).with_values(None).to_db() == json.dumps([{'id':1, 'name':'test'}]).encode('utf-8')


def test_cassandra_json_array_with_bytes():
    compare(
        actual=CassandraArray(SampleDict).with_values(
            json.dumps([{'id': 234, 'name': 'val'}]).encode('utf-8')
        ).value,
        expected=[{
            SampleDict.Id.name: 234,
            SampleDict.Name.name: 'val'
        }]
    )


def test_cassandra_json_dict_to_db():
    class CustomDict(CassandraDict):
        Id = Int(name='id')
        Name = String(name='name')

    assert CustomDict().with_values({
        CustomDict.Id.name: 1,
        CustomDict.Name.name: 'test'
    }).with_values(None).to_db() == json.dumps({'id':1, 'name':'test'}).encode('utf-8')


def test_cassandra_json_dict_with_bytes():
    class CustomDict(CassandraDict):
        Id = Int(name='id')
        Name = String(name='name')

    compare(
        actual=CustomDict().with_values(
            json.dumps({'id': 234, 'name': 'val'}).encode('utf-8')
        ).value,
        expected={
            CustomDict.Id.name: 234,
            CustomDict.Name.name: 'val'
        }
    )



import pytest

from surety.db.client import SqliteDBClient


def test_trunc_not_implemented():
    client = SqliteDBClient(':memory:')
    with pytest.raises(NotImplementedError):
        client.trunc_all_tables()


def test_context_manager(tmp_path):
    db_file = tmp_path / 'test.db'

    with SqliteDBClient(str(db_file)) as client:
        assert client.connection is not None


def test_read_write(tmp_path):
    db_file = tmp_path / 't.db'

    client = SqliteDBClient(str(db_file))
    client.connect()
    client.write_data('CREATE TABLE test(id int)')
    client.write_data('INSERT INTO test VALUES (?)', 1)

    result = client.read_data('SELECT * FROM test')
    assert result[0][0] == 1



import pytest

from surety.db import Database
from surety.db.types import DbModel

# pylint: disable=redefined-outer-name,protected-access,attribute-defined-outside-init

class FakeModel:
    __table__ = object()

    @classmethod
    def get_table_name(cls):
        return 'users'

    def __init__(self, is_custom=False):
        pass

    def with_values(self, _):
        return self


class FakeClient:
    def __init__(self):
        self.records = []

    def get_records(self, *_, **__):
        return self.records

    def insert(self, record):
        return record

    def pre_insert(self, record):
        pass

    def commit_and_close(self):
        pass

    def delete(self, *_, **__):
        self.deleted = True

    def update(self, *_, **__):
        self.updated = True

    def cascade_delete(self, _):
        self.cascade = True

    def trunc_all_tables(self, **__):
        self.truncated = True


@pytest.fixture(autouse=True)
def mock_wait(monkeypatch):
    def fake_wait(func, **__):
        return func()

    monkeypatch.setattr('surety.db.database.wait', fake_wait)


@pytest.fixture(autouse=True)
def mock_compare(monkeypatch):
    monkeypatch.setattr('surety.db.database.compare', lambda **kwargs: None)


@pytest.fixture
def db(monkeypatch):
    fake = FakeClient()
    monkeypatch.setattr('surety.db.database.DbClient', lambda **kw: fake)
    return Database('test', client=None)


def test_db_mode_toggle(db):
    db.enable_no_db_mode()
    assert db._no_db_mode is True

    db.reset_mode()
    assert db._no_db_mode is False


def test_insert_no_db_mode(db):
    db.enable_no_db_mode()
    assert db.insert('record') is None


def test_insert_records_no_db_mode(db):
    db.enable_no_db_mode()
    db.insert_records([object()])  # should early-return


def test_insert_records(db):
    class Record:
        def to_table_record(self):
            return self

    db.insert_records([Record()])


def test_delete_calls_client(db):
    db.delete('model', id=1)


def test_update_calls_client(db):
    db.update('model', {'name': 'new'}, id=1)


def test_cascade_delete(db):
    db.cascade_delete('model')


def test_verify_no_record(db):
    db.verify_no_record(type(
        'M', (), {'__table__': None, 'get_table_name': lambda: 'users'}
    ))


def test_verify_records_empty():
    Database.verify_records([], [])


def test_trunc_all_tables(db):
    db.trunc_all_tables()


def test_get_records_with_wait(db):
    db.db.records = [{'id': 1}]
    result = db.get_records(FakeModel, id=1, sleep_seconds=0)

    assert len(result) == 1


def test_get_record_wrapper(db, monkeypatch):
    monkeypatch.setattr(db, 'get_records', lambda *a, **k: [1])
    assert db.get_record(object) == 1


def test_verify_record():
    class M(DbModel):
        value = {'a': 1}

    Database.verify_record(M(), M())


def test_verify_records_non_empty():
    class M(DbModel):
        value = {'a': 1}

        @classmethod
        def get_table_name(cls):
            return 'users'

    Database.verify_records([M()], [M()])


def test_verify_no_record_asserts(db):
    db.db.records = [1]

    with pytest.raises(AssertionError):
        db.verify_no_record(FakeModel)


def test_verify_records_expected_only():
    class M:
        value = {'a': 1}

        @classmethod
        def get_table_name(cls):
            return 'users'

    Database.verify_records([], [M()])

def test_get_records_nowait(db):
    db.db.records = [{'id': 1}]

    result = db.get_records_nowait(FakeModel)
    assert len(result) == 1


def test_insert_records_real(db):
    class Record:
        def to_table_record(self):
            return self

    db.insert(Record())

def test_verify_no_record_with_wait(db, monkeypatch):
    class M(DbModel):
        value = {'a': 1}

        @classmethod
        def get_table_name(cls):
            return 'users'

    monkeypatch.setattr(
        db.db,
        'get_records',
        lambda *a, **k: []
    )

    db.verify_no_record_with_wait(M)



import json

from surety.db.types.common import DBBaseArray, DBBaseJson


class DBJsonDict(DBBaseJson):
    def to_db(self):
        return json.dumps(self.value).encode('utf-8')


class DBJsonArray(DBBaseArray):
    def to_db(self):
        return json.dumps(self.value).encode('utf-8')



import json

from surety import Bool
from surety.db.types.common import DBBaseArray, DBBaseJson

# pylint: disable=attribute-defined-outside-init

class DbBool(Bool):
    def to_db(self):
        return int(self.value)

    def with_values(self, values):
        if isinstance(values, int):
            values = bool(values)

        self._value = values

        return self


class DBJsonDict(DBBaseJson):
    def to_db(self):
        return json.dumps(self.value, separators=(',', ':'))


class DbJsonArray(DBBaseArray):
    def to_db(self):
        return json.dumps(self.value, separators=(',', ':'))



from sqlalchemy import BOOLEAN, INT, TEXT, TIMESTAMP, Column
from sqlalchemy.ext.declarative import declarative_base

from surety import Bool, Int, String
from surety.db.types.common import DbTimestamp
from surety.db.types.model import DbModel
from surety.db.types.mysql import DBJsonDict

Base = declarative_base()


class Table(Base):
    __tablename__ = 'table_name'

    id = Column(INT, primary_key=True)
    name = Column(TEXT, nullable=False)
    comment = Column(TEXT)
    is_global = Column(BOOLEAN)
    metadata_column = Column(TEXT)
    created_at = Column(TIMESTAMP)


class Model(DbModel):
    __table__ = Table

    Id = Int(name='id')
    Name = String(name='name')
    Comment = String(name='comment', required=False)
    IsGlobal = Bool(name='global')
    Metadata = String(name='metadata')
    CreatedAt = DbTimestamp(name='created_at', required=False)


class SampleDict(DBJsonDict):
    Id = Int(name='id')
    Name = String(name='name')



import pytest

from surety.diff import compare
from surety.diff.rules import has_some_value
from surety.db.types.model import CassandraDbModel
from tests.data import Model, Table

# pylint: disable=protected-access

class CassandraColumn:
    column = type('col', (), {'required': False})


class CassandraTable:
    __table_name__ = 'cassandra_table'

    id = CassandraColumn()


class CassandraModel(CassandraDbModel): # pylint: disable=abstract-method
    __table__ = CassandraTable


def test_model():
    assert Model.get_table_name() == 'table_name'


def test_regular_data():
    compare(
        actual=Model().to_db(),
        expected={
            'id': 1, 'is_global': True, 'metadata_column': 'm', 'name': 'name'
        },
        rules={
            'id': has_some_value,
            'is_global': has_some_value,
            'metadata_column': has_some_value,
            'name': has_some_value
        }
    )


def test_custom_data():
    compare(
        actual=Model(is_custom=True).to_db(),
        expected={'id': 1, 'name': 'name'},
        rules={
            'id': has_some_value,
            'name': has_some_value
        }
    )


def test_full_data():
    compare(
        actual=Model(is_full=True).to_db(),
        expected={
            'id': 1,
            'is_global': True,
            'metadata_column': 'm',
            'name': 'name',
            'comment': 's',
            'created_at': 'test'
        },
        rules={
            'id': has_some_value,
            'is_global': has_some_value,
            'metadata_column': has_some_value,
            'name': has_some_value,
            'comment': has_some_value,
            'created_at': has_some_value
        }
    )


def test_to_table_record():
    assert isinstance(Model().to_table_record(), Table)


def test_is_id_column_required():
    assert Model()._is_column_required('id') is True


def test_is_global_column_required():
    assert Model()._is_column_required('global') is False


def test_is_metadata_column_required():
    assert Model()._is_column_required('metadata') is False


def test_cassandra_required_false():
    m = CassandraModel()
    assert m._is_column_required('id') is False


def test_cassandra_get_table_name_exact():
    assert CassandraModel.get_table_name() == 'cassandra_table'


def test_cassandra_to_record_is_not_implemented():
    with pytest.raises(NotImplementedError):
        CassandraModel().to_table_record()



from surety.db.client import MysqlDBClient


def test_mysql_truncate():
    class FakeConnectionTruncate:
        def execute(self, _):
            return [('TRUNCATE test.users;',)]


    client = MysqlDBClient()
    client.connection = FakeConnectionTruncate()
    client.execute = lambda sql: None

    client.trunc_all_tables(schemas=['test'])


def test_mysql_connect_initializes_connection(monkeypatch):
    created = {}

    # Mock create_engine
    def fake_create_engine(url):
        created['engine_url'] = url
        return 'fake_engine'

    # Mock sessionmaker
    def fake_sessionmaker(bind):
        created['bind'] = bind
        return lambda: 'fake_session'

    monkeypatch.setattr(
        'surety.db.client.create_engine',
        fake_create_engine
    )

    monkeypatch.setattr(
        'surety.db.client.sessionmaker',
        fake_sessionmaker
    )

    client = MysqlDBClient(password='pwd')
    result = client.connect()

    assert result is client
    assert client.connection == 'fake_session'
    assert created['bind'] == 'fake_engine'
